
#include <sys/types.h>
#include <netinet/in.h>
#include <malloc.h>
#include <string.h>

#include <ucd-snmp-config.h>
#include <asn1.h>
#include <snmp_api.h>
#include <snmp_impl.h>
#include <snmp.h>
#include <default_store.h>


static oid default_enterprise[] = {1, 3, 6, 1, 4, 1, 3, 1, 1};

#define DEFAULT_ENTERPRISE default_enterprise
#define DEFAULT_TIME	0

#ifdef USE_REVERSE_ASNENCODING

/* on error, returns NULL (likely an encoding problem). */
u_char *
fsnmp_pdu_rbuild (struct snmp_pdu *pdu, u_char *cp, size_t *out_length)
{
#define VPCACHE_SIZE 50
  struct variable_list *vpcache[VPCACHE_SIZE];
  struct variable_list *vp, *tmpvp;
  struct sockaddr_in *pduIp = (struct sockaddr_in *)&(pdu->agent_addr);
  u_char *startcp = cp;
  int i, wrapped=0, notdone, final;

  /* DEBUGMSGTL(("snmp_pdu_rbuild","starting\n")); */
  for(vp = pdu->variables, i = VPCACHE_SIZE-1; vp;
      vp = vp->next_variable, i--) {
      if (i < 0) {
          wrapped = notdone = 1;
          i = VPCACHE_SIZE-1;
          /* DEBUGMSGTL(("snmp_pdu_rbuild","wrapped\n")); */
      }
      vpcache[i] = vp;
  }
  final = i + 1;

  do {
      for(i = final; i < VPCACHE_SIZE; i++) {
          vp = vpcache[i];
          /* DEBUGDUMPSECTION("send", "VarBind"); */
          cp = snmp_rbuild_var_op(cp, vp->name, &vp->name_length, vp->type,
                                  vp->val_len, (u_char *)vp->val.string,
                                  out_length);
          /* DEBUGINDENTLESS(); */
          if (cp == NULL)
              return NULL;
      }
      /* DEBUGINDENTLESS(); */
      if (wrapped) {
          notdone = 1;
          for(i = 0; i < final; i++) {
              vp = vpcache[i];
              /* DEBUGDUMPSECTION("send", "VarBind"); */
              cp = snmp_rbuild_var_op(cp, vp->name, &vp->name_length, vp->type,
                                      vp->val_len, (u_char *)vp->val.string,
                                      out_length);
              /* DEBUGINDENTLESS(); */
              if (cp == NULL)
                  return NULL;
          }
          
          if (final == 0)
              tmpvp = vpcache[VPCACHE_SIZE-1];
          else
              tmpvp = vpcache[final-1];
          wrapped = 0;
          for(vp = pdu->variables, i = VPCACHE_SIZE-1; vp && vp != tmpvp;
              vp = vp->next_variable, i--) {
              if (i < 0) {
                  wrapped = 1;
                  i = VPCACHE_SIZE-1;
                  /* DEBUGMSGTL(("snmp_pdu_rbuild","wrapped\n")); */
              }
              vpcache[i] = vp;
          }
          final = i + 1;
      } else {
          notdone = 0;
      }
  } while (notdone);

  /* Save current location and build SEQUENCE tag and length placeholder
       for variable-bindings sequence
       (actual length will be inserted later) */
  cp = asn_rbuild_sequence(cp, out_length,
                           (u_char)(ASN_SEQUENCE | ASN_CONSTRUCTOR),
                           startcp - cp);

  /* store fields in the PDU preceeding the variable-bindings sequence */
  if (pdu->command != SNMP_MSG_TRAP){
      /* PDU is not an SNMPv1 trap */

      /* error index (getbulk max-repetitions) */
      cp = asn_rbuild_int(cp, out_length,
                          (u_char)(ASN_UNIVERSAL | ASN_PRIMITIVE | ASN_INTEGER),
                          &pdu->errindex, sizeof(pdu->errindex));
      if (cp == NULL)
          return NULL;

      /* error status (getbulk non-repeaters) */
      cp = asn_rbuild_int(cp, out_length,
                          (u_char)(ASN_UNIVERSAL | ASN_PRIMITIVE | ASN_INTEGER),
                          &pdu->errstat, sizeof(pdu->errstat));
      if (cp == NULL)
          return NULL;


      /* request id */
      cp = asn_rbuild_int(cp, out_length,
                          (u_char)(ASN_UNIVERSAL | ASN_PRIMITIVE | ASN_INTEGER),
                          &pdu->reqid, sizeof(pdu->reqid));
      if (cp == NULL)
          return NULL;

  } else {
    /* an SNMPv1 trap PDU */

      /* timestamp  */
      cp = asn_rbuild_unsigned_int(cp, out_length,
                                  (u_char)(ASN_TIMETICKS | ASN_PRIMITIVE),
                                  &pdu->time, sizeof(pdu->time));
      if (cp == NULL)
          return NULL;

      /* specific trap */
      cp = asn_rbuild_int(cp, out_length,
                         (u_char)(ASN_UNIVERSAL | ASN_PRIMITIVE | ASN_INTEGER),
                         (long *)&pdu->specific_type,
                         sizeof(pdu->specific_type));
      if (cp == NULL)
          return NULL;

      /* generic trap */
      cp = asn_rbuild_int(cp, out_length,
                         (u_char)(ASN_UNIVERSAL | ASN_PRIMITIVE | ASN_INTEGER),
                         (long *)&pdu->trap_type, sizeof(pdu->trap_type));
      if (cp == NULL)
          return NULL;


      /* agent-addr */
      cp = asn_rbuild_string(cp, out_length,
                            (u_char)(ASN_IPADDRESS | ASN_PRIMITIVE),
                            (u_char *)&pduIp->sin_addr.s_addr,
                            sizeof(pduIp->sin_addr.s_addr));
      if (cp == NULL)
          return NULL;


      /* enterprise */
      cp = asn_rbuild_objid(cp, out_length,
                           (u_char)
                           (ASN_UNIVERSAL | ASN_PRIMITIVE | ASN_OBJECT_ID),
                           (oid *)pdu->enterprise, pdu->enterprise_length);
      if (cp == NULL)
          return NULL;

  }

  /* build the PDU sequence */
  cp = asn_rbuild_sequence(cp, out_length,
                           (u_char)pdu->command,
                           startcp - cp);

  return cp;
}
#endif /* USE_REVERSE_ASNENCODING */


/*
 * Takes a session and a pdu and serializes the ASN PDU into the area
 * pointed to by packet.  out_length is the size of the data area available.
 * Returns the length of the completed packet in out_length.  If any errors
 * occur, -1 is returned.  If all goes well, 0 is returned.
 */

u_char *fsnmp_build_packet(struct snmp_pdu *pdu,
             u_char *packet,
             size_t *out_length,
	     int *eno)
{
    u_char *h0, *h0e = 0, *h1;
    u_char  *cp;
    size_t length;
    long version;

    *eno=0;

    switch (pdu->command) {
	case SNMP_MSG_RESPONSE:
	    pdu->flags &= (~UCD_MSG_FLAG_EXPECT_RESPONSE);
		/* Fallthrough */
	case SNMP_MSG_GET:
	case SNMP_MSG_GETNEXT:
	case SNMP_MSG_SET:
            /* all versions support these PDU types */
            /* initialize defaulted PDU fields */

	    if (pdu->errstat == SNMP_DEFAULT_ERRSTAT)
	        pdu->errstat = 0;
	    if (pdu->errindex == SNMP_DEFAULT_ERRINDEX)
	        pdu->errindex = 0;
	    break;

	case SNMP_MSG_TRAP2:
	    pdu->flags &= (~UCD_MSG_FLAG_EXPECT_RESPONSE);
		/* Fallthrough */
	case SNMP_MSG_INFORM:
            /* not supported in SNMPv1 and SNMPsec */
	    if (pdu->version == SNMP_VERSION_1) {
		    *eno=SNMPERR_V2_IN_V1;
	            return(NULL);
	    }
	    if (pdu->errstat == SNMP_DEFAULT_ERRSTAT)
	        pdu->errstat = 0;
	    if (pdu->errindex == SNMP_DEFAULT_ERRINDEX)
	        pdu->errindex = 0;
	    break;

	case SNMP_MSG_GETBULK:
            /* not supported in SNMPv1 and SNMPsec */
	    if (pdu->version == SNMP_VERSION_1) {
		    *eno=SNMPERR_V2_IN_V1;
	            return(NULL);
            }
	    if (pdu->max_repetitions < 0) {
		*eno=SNMPERR_V2_IN_V1;
	        return(NULL);
	    }
	    if (pdu->non_repeaters < 0) {
		*eno=SNMPERR_BAD_REPEATERS;
	        return(NULL);
	    }
	    break;

	case SNMP_MSG_TRAP:
            /* *only* supported in SNMPv1 and SNMPsec */
	    if (pdu->version != SNMP_VERSION_1) {
		    *eno=SNMPERR_V1_IN_V2;
	            return(NULL);
            }
            /* initialize defaulted Trap PDU fields */
	    pdu->reqid = 1;	/* give a bogus non-error reqid for traps */
	    if (pdu->enterprise_length == SNMP_DEFAULT_ENTERPRISE_LENGTH){
	        pdu->enterprise = (oid *)malloc(sizeof(DEFAULT_ENTERPRISE));
	        if (pdu->enterprise == NULL) {
		    *eno=SNMPERR_MALLOC;
	            return(NULL);
	        }
	        memmove(pdu->enterprise, DEFAULT_ENTERPRISE,
		    sizeof(DEFAULT_ENTERPRISE));
	        pdu->enterprise_length = sizeof(DEFAULT_ENTERPRISE)/sizeof(oid);
	    }
	    if (pdu->time == SNMP_DEFAULT_TIME)
	        pdu->time = DEFAULT_TIME;
            /* don't expect a response */
	    pdu->flags &= (~UCD_MSG_FLAG_EXPECT_RESPONSE);
	    break;

	case SNMP_MSG_REPORT:		/* SNMPv3 only */
	default:
	    *eno=SNMPERR_UNKNOWN_PDU;
            return(NULL);
    }

    /* save length */
    length = *out_length;
    /* setup administrative fields based on version */
    /* build the message wrapper and all the administrative fields
       upto the PDU sequence
       (note that actual length of message will be inserted later) */
    h0 = packet;
    switch (pdu->version) {
    case SNMP_VERSION_1:
    case SNMP_VERSION_2c:
	if (pdu->community_len == 0 || 
		pdu->community == NULL) {
		*eno = SNMPERR_BAD_COMMUNITY;
		return(NULL);
	}

#ifdef USE_REVERSE_ASNENCODING
        if (ds_get_boolean(DS_LIBRARY_ID, DS_LIB_REVERSE_ENCODE)) {
	    packet=packet+*out_length-1;
            cp = fsnmp_pdu_rbuild(pdu, packet, out_length);
            if (cp == NULL)
                return(NULL);

            cp = asn_rbuild_string(cp, out_length,
                                   (u_char)
                                   (ASN_UNIVERSAL | ASN_PRIMITIVE | ASN_OCTET_STR),
                                   pdu->community, pdu->community_len);
            if (cp == NULL)
                return(NULL);

        
        /* store the version field */

            version = pdu->version;
            cp = asn_rbuild_int(cp, out_length,
                                (u_char)
                                (ASN_UNIVERSAL | ASN_PRIMITIVE | ASN_INTEGER),
                                (long *) &version, sizeof(version));
            if (cp == NULL)
                return(NULL);

#if 0
        /* build the final sequence */
            if (pdu->version == SNMP_VERSION_1)
                DEBUGDUMPSECTION("send", "SNMPv1 Message");
            else
                DEBUGDUMPSECTION("send", "SNMPv2c Message");
#endif
            cp = asn_rbuild_sequence(cp, out_length,
                                     (u_char)(ASN_SEQUENCE | ASN_CONSTRUCTOR),
                                     packet - cp);

            if (cp == NULL)
                return(NULL);

	    *out_length=packet-cp;
	    cp++;

            return(cp);
        } else {

#endif /* USE_REVERSE_ASNENCODING */
            /* Save current location and build SEQUENCE tag and length
               placeholder for SNMP message sequence
               (actual length will be inserted later) */
            cp = asn_build_sequence(packet, out_length,
                                    (u_char)(ASN_SEQUENCE | ASN_CONSTRUCTOR),
                                    0);
            if (cp == NULL)
                return(NULL);
            h0e = cp;
#if 0
            if (pdu->version == SNMP_VERSION_1)
                DEBUGDUMPSECTION("send", "SNMPv1 Message");
            else
                DEBUGDUMPSECTION("send", "SNMPv2c Message");
        
            /* store the version field */
            DEBUGDUMPHEADER("send", "SNMP Version Number");
#endif
            version = pdu->version;
            cp = asn_build_int(cp, out_length,
                               (u_char)
                               (ASN_UNIVERSAL | ASN_PRIMITIVE | ASN_INTEGER),
                               (long *) &version, sizeof(version));
            /* DEBUGINDENTLESS(); */
            if (cp == NULL)
                return(NULL);

            /* store the community string */
            /* DEBUGDUMPHEADER("send", "Community String"); */
            cp = asn_build_string(cp, out_length,
                                  (u_char) (ASN_UNIVERSAL | ASN_PRIMITIVE |
                                            ASN_OCTET_STR),
                                  pdu->community, pdu->community_len);
            /* DEBUGINDENTLESS(); */
            if (cp == NULL)
                return(NULL);
            break;

#ifdef USE_REVERSE_ASNENCODING
        }
#endif /* USE_REVERSE_ASNENCODING */

    case SNMP_VERSION_2p:
    case SNMP_VERSION_sec:
    case SNMP_VERSION_2u:
    case SNMP_VERSION_2star:
    default:
	*eno=SNMPERR_BAD_VERSION;
	return(NULL);
    }

    h1 = cp;
    /* DEBUGPRINTPDUTYPE("send", pdu->command); */
    cp = snmp_pdu_build(pdu, cp, out_length);
    /* DEBUGINDENTADD(-4); */ /* return from entire v1/v2c message */ 
    if (cp == NULL)
	return(NULL);

    /* insert the actual length of the message sequence */
    switch (pdu->version) {
    case SNMP_VERSION_1:
    case SNMP_VERSION_2c:
        asn_build_sequence(packet, &length,
		       (u_char)(ASN_SEQUENCE | ASN_CONSTRUCTOR),
                       cp - h0e);
        break;

    case SNMP_VERSION_2p:
    case SNMP_VERSION_sec:
    case SNMP_VERSION_2u:
    case SNMP_VERSION_2star:
    default:
	*eno=SNMPERR_BAD_VERSION;
	return(NULL);
    }
    *out_length = cp - packet;
    return(packet);
}
